import warnings
from typing import Any, Callable

import pyro
import torch

from chirho.robust.internals.utils import pytree_generalized_manual_revjvp
from chirho.robust.ops import Functional, P, Point, S, T, influence_fn


class MonteCarloInfluenceEstimator(pyro.poutine.messenger.Messenger):
    """
    Effect handler for approximating efficient influence functions with nested monte carlo.
    See the MC-EIF estimator in https://arxiv.org/pdf/2403.00158.pdf for details and
    :func:`~chirho.robust.ops.influence_fn` for example usage.

    .. note::

        * ``functional`` must compose with ``torch.func.jvp``
        * Since the efficient influence function is approximated using Monte Carlo, the result
          of this function is stochastic, i.e., evaluating this function on the same ``points``
          can result in different values. To reduce variance, increase ``num_samples_outer`` and
          ``num_samples_inner`` in ``linearize_kwargs``.
        * Currently, ``model`` cannot contain any ``pyro.param`` statements.
          This issue will be addressed in a future release:
          https://github.com/BasisResearch/chirho/issues/393.
        * There are memory leaks when calling this function multiple times due to ``torch.func``.
          See issue:
          https://github.com/BasisResearch/chirho/issues/516.
          To avoid this issue, use ``torch.no_grad()`` as shown in the example above.

    """

    def __init__(self, **linearize_kwargs):
        self.linearize_kwargs = linearize_kwargs
        super().__init__()

    def _pyro_influence(self, msg) -> None:
        models = msg["kwargs"]["models"]
        functional = msg["kwargs"]["functional"]
        points = msg["kwargs"]["points"]
        pointwise_influence = msg["kwargs"]["pointwise_influence"]

        args = msg["args"]
        kwargs = {
            k: v
            for k, v in msg["kwargs"].items()
            if k not in ["models", "functional", "points", "pointwise_influence"]
        }

        if len(points) != 1:
            raise NotImplementedError(
                "MonteCarloInfluenceEstimator currently only supports unary functionals"
            )

        from chirho.robust.internals.linearize import linearize
        from chirho.robust.internals.utils import make_functional_call

        if len(models) != len(points):
            raise ValueError("mismatch between number of models and points")

        linearized = linearize(
            pointwise_influence=pointwise_influence, *models, **self.linearize_kwargs
        )
        target = functional(*models)

        # TODO check that target_params == model_params
        assert isinstance(target, torch.nn.Module)
        target_params, func_target = make_functional_call(target)

        if torch.is_grad_enabled():
            warnings.warn(
                "Calling influence_fn with torch.grad enabled can lead to memory leaks. "
                "Please use torch.no_grad() to avoid this issue. See example in the docstring."
            )
        param_eif = linearized(*points, *args, **kwargs)

        # Compute the jacobian vector product of the functionals jacobian wrt the parameters with the fisher matrix X
        #  log prob of data product. This implementation uses reverse mode auto diff for the jacobian and then
        #  manually right multiplies with param_eif. Unfortunately, torch.func.jvp uses very large amounts of memory,
        #  which is exacerbated when using vmap to broadcast, as opposed to the manual impplementation used herein.
        msg["value"] = pytree_generalized_manual_revjvp(
            fn=lambda p: func_target(p, *args, **kwargs),
            params=target_params,
            batched_vector=param_eif,
        )

        msg["done"] = True


def one_step_corrected_estimator(
    functional: Functional[P, S],
    *test_points: Point[T],
) -> Functional[P, S]:
    """
    Returns a functional that computes the one-step correction for the
    functional at a specified set of test points as discussed in [1].

    :param functional: model summary functional of interest
    :param test_points: points at which to compute the one-step correction
    :return: functional to compute the one-step correction

    **References**

    [1] `Semiparametric doubly robust targeted double machine learning: a review`,
    Edward H. Kennedy, 2022.
    """
    eif_fn = influence_fn(functional, *test_points, pointwise_influence=False)

    def _corrected_functional(*model: Callable[P, Any]) -> Callable[P, S]:
        plug_in_estimator = functional(*model)
        correction_estimator = eif_fn(*model)

        def _estimator(*args, **kwargs) -> S:
            plug_in_estimate = plug_in_estimator(*args, **kwargs)
            correction = correction_estimator(*args, **kwargs)

            flat_plug_in_estimate, treespec = torch.utils._pytree.tree_flatten(
                plug_in_estimate
            )
            flat_correction, _ = torch.utils._pytree.tree_flatten(correction)

            return torch.utils._pytree.tree_unflatten(
                [a + b for a, b in zip(flat_plug_in_estimate, flat_correction)],
                treespec,
            )

        return _estimator

    return _corrected_functional



from __future__ import annotations

import heapq
import math
import typing
import warnings
from typing import Callable, Generic, List, Optional, Tuple, TypeVar, Union

import pyro
import torch

from chirho.dynamical.internals._utils import Prioritized, ShallowMessenger
from chirho.dynamical.ops import Dynamics, State, on

R = Union[float, torch.Tensor]
S = TypeVar("S")
T = TypeVar("T")


class Interruption(Generic[T], ShallowMessenger):
    predicate: Callable[[State[T]], bool]
    callback: Callable[[Dynamics[T], State[T]], Tuple[Dynamics[T], State[T]]]

    def __init__(
        self,
        predicate: Callable[[State[T]], bool],
        callback: Callable[[Dynamics[T], State[T]], Tuple[Dynamics[T], State[T]]],
    ):
        self.predicate = predicate
        self.callback = callback

    def _pyro_get_new_interruptions(self, msg: dict) -> None:
        if msg["value"] is None:
            msg["value"] = []
        assert isinstance(msg["value"], list)
        msg["value"].append(self)


@pyro.poutine.runtime.effectful(type="get_new_interruptions")
def get_new_interruptions() -> List[Interruption]:
    """
    Install the active interruptions into the context.
    """
    return []


class Solver(Generic[T], pyro.poutine.messenger.Messenger):
    @staticmethod
    def _prioritize_interruption(h: Interruption[T]) -> Prioritized[Interruption[T]]:
        from chirho.dynamical.handlers.interruption import StaticEvent, ZeroEvent

        if isinstance(h.predicate, StaticEvent):
            return Prioritized(float(h.predicate.time), h)
        elif isinstance(h.predicate, ZeroEvent):
            return Prioritized(-math.inf, h)
        else:
            raise NotImplementedError(f"cannot install interruption {h}")

    @typing.final
    def _pyro_simulate(self, msg: dict) -> None:
        from chirho.dynamical.handlers.interruption import StaticEvent

        dynamics: Dynamics[T] = msg["args"][0]
        state: State[T] = msg["args"][1]
        start_time: R = msg["args"][2]
        end_time: R = msg["args"][3]

        # needed to avoid mypy
        assert isinstance(start_time, torch.Tensor)
        assert isinstance(end_time, torch.Tensor)

        if pyro.settings.get("validate_dynamics"):
            check_dynamics(dynamics, state, start_time, end_time, **msg["kwargs"])

        # local state
        all_interruptions: List[Prioritized[Interruption[T]]] = []
        heapq.heappush(
            all_interruptions,
            self._prioritize_interruption(
                on(StaticEvent(end_time), lambda d, s: (d, s))
            ),
        )

        while start_time < end_time:
            for h in get_new_interruptions():
                if isinstance(h.predicate, StaticEvent) and h.predicate.time > end_time:
                    warnings.warn(
                        f"{Interruption.__name__} {h} with time={h.predicate.time} "
                        f"occurred after the end of the timespan ({start_time}, {end_time})."
                        "This interruption will have no effect.",
                        UserWarning,
                    )
                elif (
                    isinstance(h.predicate, StaticEvent)
                    and h.predicate.time < start_time
                ):
                    raise ValueError(
                        f"{Interruption.__name__} {h} with time {h.predicate.time} "
                        f"occurred before the start of the timespan ({start_time}, {end_time})."
                        "This interruption will have no effect."
                    )
                else:
                    heapq.heappush(all_interruptions, self._prioritize_interruption(h))

            possible_interruptions: List[Interruption[T]] = []
            while all_interruptions:
                ph: Prioritized[Interruption[T]] = heapq.heappop(all_interruptions)
                possible_interruptions.append(ph.item)
                if ph.priority > start_time:
                    break

            next_interruption: Optional[Interruption[T]]
            state, start_time, next_interruption = simulate_to_interruption(
                possible_interruptions,
                dynamics,
                state,
                start_time,
                end_time,
                **msg["kwargs"],
            )

            if next_interruption is not None:
                # TODO reenable this check once we figure out why mypy is inferring "Never"s
                dynamics, state = next_interruption.callback(dynamics, state)  # type: ignore

                for h in possible_interruptions:
                    if h is not next_interruption:
                        heapq.heappush(
                            all_interruptions, self._prioritize_interruption(h)
                        )

        msg["value"] = state
        msg["done"] = True


@pyro.poutine.runtime.effectful(type="simulate_point")
def simulate_point(
    dynamics: Dynamics[T],
    initial_state: State[T],
    start_time: R,
    end_time: R,
    **kwargs,
) -> State[T]:
    """
    Simulate a dynamical system.
    """
    raise NotImplementedError("No default behavior for simulate_point")


@pyro.poutine.runtime.effectful(type="simulate_trajectory")
def simulate_trajectory(
    dynamics: Dynamics[T],
    initial_state: State[T],
    timespan: R,
    **kwargs,
) -> State[T]:
    """
    Simulate a dynamical system.
    """
    raise NotImplementedError("No default behavior for simulate_trajectory")


@pyro.poutine.runtime.effectful(type="simulate_to_interruption")
def simulate_to_interruption(
    interruption_stack: List[Interruption[T]],
    dynamics: Dynamics[T],
    start_state: State[T],
    start_time: R,
    end_time: R,
    **kwargs,
) -> Tuple[State[T], R, Optional[Interruption[T]]]:
    """
    Simulate a dynamical system until the next interruption.

    :returns: the final state
    """
    if len(interruption_stack) == 0:
        return (
            simulate_point(dynamics, start_state, start_time, end_time, **kwargs),
            end_time,
            None,
        )

    raise NotImplementedError("No default behavior for simulate_to_interruption")


@pyro.poutine.runtime.effectful(type="check_dynamics")
def check_dynamics(
    dynamics: Dynamics[T],
    initial_state: State[T],
    start_time: R,
    end_time: R,
    **kwargs,
) -> None:
    """
    Validate a dynamical system.
    """
    pass


DYNAMICS_VALIDATION_ENABLED: bool = False


@pyro.settings.register("validate_dynamics", __name__, "DYNAMICS_VALIDATION_ENABLED")
def _check_validate_dynamics_flag(value: bool) -> None:
    assert isinstance(value, bool)



import sys

from setuptools import find_packages, setup

VERSION = "0.3.0"

try:
    long_description = open("README.rst", encoding="utf-8").read()
except Exception as e:
    sys.stderr.write("Failed to read README: {}\n".format(e))
    sys.stderr.flush()
    long_description = ""

# examples/tutorials
EXTRAS_REQUIRE = [
    "jupyter",
    "graphviz",
    "matplotlib",
    "pandas",
    "seaborn",
    "pytorch-lightning",
    "scikit-image",
    "tensorboard",
    "typing_extensions",
]

DYNAMICAL_REQUIRE = ["torchdiffeq"]

setup(
    name="chirho",
    version=VERSION,
    description="Causal reasoning",
    long_description=long_description,
    packages=find_packages(include=["chirho", "chirho.*"]),
    author="Basis",
    url="https://www.basis.ai/",
    project_urls={
    #     "Documentation": "",
        "Source": "https://github.com/BasisResearch/chirho",
    },
    install_requires=[
        # if you add any additional libraries, please also
        # add them to `docs/source/requirements.txt`
        "pyro-ppl>=1.9.1",
    ],
    extras_require={
        "dynamical": DYNAMICAL_REQUIRE,
        "extras": EXTRAS_REQUIRE,
        "test": EXTRAS_REQUIRE + DYNAMICAL_REQUIRE
        + [
            "pytest",
            "pytest-cov",
            "pytest-xdist",
            "mypy==1.10",
            "black",
            "flake8",
            "isort",
            "sphinx==7.1.2",
            "sphinxcontrib-bibtex",
            "sphinxcontrib-jquery",
            "sphinx_rtd_theme==1.3.0",
            "myst_parser",
            "nbsphinx",
            "nbval",
            "nbqa",
        ],
    },
    python_requires=">=3.9",
    keywords="machine learning statistics probabilistic programming bayesian modeling pytorch",
    license="Apache 2.0",
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",

        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    # yapf
)



import contextlib
from typing import Callable, Mapping, Optional, TypeVar, Union, cast

import pyro.distributions.constraints as constraints
import torch

from chirho.explainable.handlers.components import (
    consequent_eq_neq,
    random_intervention,
    sufficiency_intervention,
    undo_split,
)
from chirho.explainable.handlers.preemptions import Preemptions
from chirho.interventional.handlers import do
from chirho.interventional.ops import Intervention
from chirho.observational.handlers.condition import Factors
from chirho.observational.ops import Observation

S = TypeVar("S")
T = TypeVar("T")


@contextlib.contextmanager
def SplitSubsets(
    supports: Mapping[str, constraints.Constraint],
    actions: Mapping[str, Intervention[T]],
    *,
    bias: float = 0.0,
    prefix: str = "__cause_split_",
):
    """
    A context manager used for a stochastic search of minimal but-for causes among potential interventions.
    On each run, nodes listed in ``actions`` are randomly selected and intervened on with probability ``.5 + bias``
    (that is, preempted with probability ``.5-bias``). The sampling is achieved by adding stochastic binary preemption
    nodes associated with intervention candidates. If a given preemption node has value ``0``, the corresponding
    intervention is executed. See tests in `tests/explainable/test_handlers_explanation.py` for examples.

    :param supports: A mapping of sites to their support constraints.
    :param actions: A mapping of sites to interventions.
    :param bias: The scalar bias towards not intervening. Must be between -0.5 and 0.5, defaults to 0.0.
    :param prefix: A prefix used for naming additional preemption nodes. Defaults to ``__cause_split_``.
    """
    preemptions = {
        antecedent: undo_split(supports[antecedent], antecedents=[antecedent])
        for antecedent in actions.keys()
    }

    with do(actions=actions):
        with Preemptions(actions=preemptions, bias=bias, prefix=prefix):
            yield


@contextlib.contextmanager
def SearchForExplanation(
    supports: Mapping[str, constraints.Constraint],
    antecedents: Mapping[str, Optional[Observation[S]]],
    consequents: Mapping[str, Optional[Observation[T]]],
    witnesses: Optional[
        Mapping[str, Optional[Union[Observation[S], Observation[T]]]]
    ] = None,
    *,
    alternatives: Optional[Mapping[str, Intervention[S]]] = None,
    factors: Optional[Mapping[str, Callable[[T], torch.Tensor]]] = None,
    preemptions: Optional[Mapping[str, Union[Intervention[S], Intervention[T]]]] = None,
    consequent_scale: float = 1e-2,
    antecedent_bias: float = 0.0,
    witness_bias: float = 0.0,
    prefix: str = "__cause__",
):
    """
    A handler for transforming causal explanation queries into probabilistic inferences.

    When used as a context manager, ``SearchForExplanation`` yields a dictionary of observations
    that can be used with ``condition`` to simultaneously impose an additional factivity constraint
    alongside the necessity and sufficiency constraints implemented by ``SearchForExplanation`` ::

        with SearchForExplanation(supports, antecedents, consequents, ...) as evidence:
            with condition(data=evidence):
                model()

    :param supports: A mapping of sites to their support constraints.
    :param antecedents: A mapping of antecedent names to optional observations.
    :param consequents: A mapping of consequent names to optional observations.
    :param witnesses: A mapping of witness names to optional observations.
    :param alternatives: An optional mapping of names to alternative antecedent interventions.
    :param factors: An optional mapping of names to consequent constraint factors.
    :param preemptions: An optional mapping of names to witness preemption values.
    :param antecedent_bias: The scalar bias towards not intervening. Must be between -0.5 and 0.5, defaults to 0.0.
    :param consequent_scale: The scale of the consequent factor functions, defaults to 1e-2.
    :param witness_bias: The scalar bias towards not preempting. Must be between -0.5 and 0.5, defaults to 0.0.
    :param prefix: A prefix used for naming additional consequent nodes. Defaults to ``__consequent_``.

    .. note:: If both antedecent and witness auxiliary nodes are provided for a site, antecedent takes priority:
    i.e., if the site is not antecedent-preempted (so the antecedent intervention is executed), it will only be
    witness-preempted in the factual world (non-trivial only if non-trivial preemptions are passed), and alternative
    interventions will still be performed in the counterfactual worlds.

    :return: A context manager that can be used to query the evidence.
    """
    ########################################
    # Validate input arguments
    ########################################
    assert len(antecedents) > 0
    assert len(consequents) > 0
    assert not set(consequents.keys()) & set(antecedents.keys())
    assert set(antecedents.keys()) <= set(supports.keys())
    assert set(consequents.keys()) <= set(supports.keys())
    if witnesses is not None:
        assert set(witnesses.keys()) <= set(supports.keys())
        assert not set(witnesses.keys()) & set(consequents.keys())
    else:
        # if witness candidates are not provided, use all non-consequent nodes
        witnesses = {w: None for w in set(supports.keys()) - set(consequents.keys())}

    ##################################################################
    # Fill in default argument values and create constituent handlers
    ##################################################################

    # defaults for necessity interventions
    alternatives = (
        {a: alternatives[a] for a in antecedents.keys()}
        if alternatives is not None
        else {
            a: cast(
                Callable,
                random_intervention(supports[a], name=f"{prefix}_alternative_{a}"),
            )
            for a in antecedents.keys()
        }
    )

    # defaults for sufficiency interventions
    sufficiency_actions = {
        a: (
            antecedents[a]
            if antecedents[a] is not None
            else sufficiency_intervention(supports[a], antecedents=antecedents.keys())
        )
        for a in antecedents.keys()
    }

    # interventions on subsets of antecedents
    antecedent_handler = SplitSubsets(
        {a: supports[a] for a in antecedents.keys()},
        {a: (alternatives[a], sufficiency_actions[a]) for a in antecedents.keys()},  # type: ignore
        bias=antecedent_bias,
        prefix=f"{prefix}__antecedent_",
    )

    witness_preemptions = {
        w: (
            preemptions[w]
            if preemptions is not None and preemptions[w] is not None
            else undo_split(supports[w], antecedents=antecedents.keys())
        )
        for w in witnesses
    }

    witness_handler = Preemptions(
        (witness_preemptions),
        bias=witness_bias,
        prefix=f"{prefix}__witness_",
    )

    consequent_handler: Factors[T] = Factors(
        (
            {c: factors[c] for c in consequents.keys()}
            if factors is not None
            else {
                c: consequent_eq_neq(
                    support=supports[c],
                    proposed_consequent=consequents[c],  # added this
                    antecedents=antecedents.keys(),
                    scale=consequent_scale,
                )
                for c in consequents.keys()
            }
        ),
        prefix=f"{prefix}__consequent_",
    )

    ######################################################################
    # Apply handlers and yield evidence for optional factual conditioning
    ######################################################################
    evidence: Mapping[str, Union[Observation[S], Observation[T]]] = {
        **{a: aa for a, aa in antecedents.items() if aa is not None},
        **{c: cc for c, cc in consequents.items() if cc is not None},
        **{w: ww for w, ww in (witnesses or {}).items() if ww is not None},
    }
    with witness_handler, antecedent_handler, consequent_handler:
        yield evidence



import functools
import operator
from typing import Dict, Iterable, List, Optional, Set, Tuple, TypeVar, Union

import pyro
import torch

T = TypeVar("T")


class IndexSet(Dict[str, Set[int]]):
    """
    :class:`IndexSet` s represent the support of an indexed value, primarily
    those created using :func:`intervene` and :class:`MultiWorldCounterfactual`
    for which free variables correspond to single interventions and indices
    to worlds where that intervention either did or did not happen.

    :class:`IndexSet` can be understood conceptually as generalizing :class:`torch.Size`
    from multidimensional arrays to arbitrary values, from positional to named dimensions,
    and from bounded integer interval supports to finite sets of positive integers.

    :class:`IndexSet` s are implemented as :class:`dict` s with
    :class:`str` s as keys corresponding to names of free index variables
    and :class:`set` s of positive :class:`int` s as values corresponding
    to the values of the index variables where the indexed value is defined.

    For example, the following :class:`IndexSet` represents
    the sets of indices of the free variables ``x`` and ``y``
    for which a value is defined::

        >>> IndexSet(x={0, 1}, y={2, 3}})
        {"x": {0, 1}, "y": {2, 3}}

    :class:`IndexSet` 's constructor will automatically drop empty entries
    and attempt to convert input values to :class:`set` s::

        >>> IndexSet(x=[0, 0, 1], y=set(), z=2)
        {"x": {0, 1}, "z": {2}}

    :class:`IndexSet` s are also hashable and can be used as keys in :class:`dict` s::

        >>> indexset = IndexSet(x={0, 1}, y={2, 3}})
        >>> indexset in {indexset: 1}
        True
    """

    def __init__(self, **mapping: Union[int, Iterable[int]]):
        super().__init__(
            **{
                k: {vs} if isinstance(vs, int) else set(vs)
                for k, vs in mapping.items()
                if vs
            }
        )

    def __repr__(self):
        return f"{type(self).__name__}({super().__repr__()})"

    def __hash__(self):
        return hash(frozenset((k, frozenset(vs)) for k, vs in self.items()))


def union(*indexsets: IndexSet) -> IndexSet:
    """
    Compute the union of multiple :class:`IndexSet` s
    as the union of their keys and of value sets at shared keys.

    If :class:`IndexSet` may be viewed as a generalization of :class:`torch.Size`,
    then :func:`union` is a generalization of :func:`torch.broadcast_shapes`
    for the more abstract :class:`IndexSet` data structure.

    Example::

        >>> union(IndexSet(a={0, 1}, b={1}), IndexSet(a={1, 2}))
        {"a": {0, 1, 2}, "b": {1}}

    .. note::

        :func:`union` satisfies several algebraic equations for arbitrary inputs.
        In particular, it is associative, commutative, idempotent and absorbing::

            union(a, union(b, c)) == union(union(a, b), c)
            union(a, b) == union(b, a)
            union(a, a) == a
            union(a, union(a, b)) == union(a, b)
    """
    return IndexSet(
        **{
            k: set.union(*[vs[k] for vs in indexsets if k in vs])
            for k in set.union(*(set(vs) for vs in indexsets))
        }
    )


@functools.singledispatch
def indices_of(value, **kwargs) -> IndexSet:
    """
    Get a :class:`IndexSet` of indices on which an indexed value is supported.
    :func:`indices_of` is useful in conjunction with :class:`MultiWorldCounterfactual`
    for identifying the worlds where an intervention happened upstream of a value.

    For example, in a model with an outcome variable ``Y`` and a treatment variable
    ``T`` that has been intervened on, ``T`` and ``Y`` are both indexed by ``"T"``::

        >>> with MultiWorldCounterfactual():
        ...     X = pyro.sample("X", get_X_dist())
        ...     T = pyro.sample("T", get_T_dist(X))
        ...     T = intervene(T, t, name="T_ax")  # adds an index variable "T_ax"
        ...     Y = pyro.sample("Y", get_Y_dist(X, T))
        ...     assert indices_of(X) == IndexSet()
        ...     assert indices_of(T) == IndexSet(T_ax={0, 1})
        ...     assert indices_of(Y) == IndexSet(T_ax={0, 1})

    Just as multidimensional arrays can be expanded to shapes with new dimensions
    over which they are constant, :func:`indices_of` is defined extensionally,
    meaning that values are treated as constant functions of free variables
    not in their support.

    .. note::

        :func:`indices_of` can be extended to new value types by registering
        an implementation for the type using :func:`functools.singledispatch` .

    .. note::

        Fully general versions of :func:`indices_of` , :func:`gather`
        and :func:`scatter` would require a dependent broadcasting semantics
        for indexed values, as is the case in sparse or masked array libraries
        like ``torch.sparse`` or relational databases.

        However, this is beyond the scope of this library as it currently exists.
        Instead, :func:`gather` currently binds free variables in its input indices
        when their indices there are a strict subset of the corresponding indices
        in ``value`` , so that they no longer appear as free in the result.

        For example, in the above snippet, applying :func:`gather` to to select only
        the values of ``Y`` from worlds where no intervention on ``T`` happened
        would result in a value that no longer contains free variable ``"T"``::

            >>> indices_of(Y) == IndexSet(T_ax={0, 1})
            True
            >>> Y0 = gather(Y, IndexSet(T_ax={0}))
            >>> indices_of(Y0) == IndexSet() != IndexSet(T_ax={0})
            True

        The practical implications of this imprecision are limited
        since we rarely need to :func:`gather` along a variable twice.

    :param value: A value.
    :param kwargs: Additional keyword arguments used by specific implementations.
    :return: A :class:`IndexSet` containing the indices on which the value is supported.
    """
    raise NotImplementedError


@functools.singledispatch
def gather(value, indexset: IndexSet, **kwargs):
    """
    Selects entries from an indexed value at the indices in a :class:`IndexSet` .
    :func:`gather` is useful in conjunction with :class:`MultiWorldCounterfactual`
    for selecting components of a value corresponding to specific counterfactual worlds.

    For example, in a model with an outcome variable ``Y`` and a treatment variable
    ``T`` that has been intervened on, we can use :func:`gather` to define quantities
    like treatment effects that require comparison of different potential outcomes::

        >>> with MultiWorldCounterfactual():
        ...     X = pyro.sample("X", get_X_dist())
        ...     T = pyro.sample("T", get_T_dist(X))
        ...     T = intervene(T, t, name="T_ax")  # adds an index variable "T_ax"
        ...     Y = pyro.sample("Y", get_Y_dist(X, T))
        ...     Y_factual = gather(Y, IndexSet(T_ax=0))         # no intervention
        ...     Y_counterfactual = gather(Y, IndexSet(T_ax=1))  # intervention
        ...     treatment_effect = Y_counterfactual - Y_factual

    Like :func:`torch.gather` and substitution in term rewriting,
    :func:`gather` is defined extensionally, meaning that values
    are treated as constant functions of variables not in their support.

    :func:`gather` will accordingly ignore variables in ``indexset``
    that are not in the support of ``value`` computed by :func:`indices_of` .

    .. note::

        :func:`gather` can be extended to new value types by registering
        an implementation for the type using :func:`functools.singledispatch` .

    .. note::

        Fully general versions of :func:`indices_of` , :func:`gather`
        and :func:`scatter` would require a dependent broadcasting semantics
        for indexed values, as is the case in sparse or masked array libraries
        like ``scipy.sparse`` or ``xarray`` or in relational databases.

        However, this is beyond the scope of this library as it currently exists.
        Instead, :func:`gather` currently binds free variables in ``indexset``
        when their indices there are a strict subset of the corresponding indices
        in ``value`` , so that they no longer appear as free in the result.

        For example, in the above snippet, applying :func:`gather` to to select only
        the values of ``Y`` from worlds where no intervention on ``T`` happened
        would result in a value that no longer contains free variable ``"T"``::

            >>> indices_of(Y) == IndexSet(T_ax={0, 1})
            True
            >>> Y0 = gather(Y, IndexSet(T_ax={0}))
            >>> indices_of(Y0) == IndexSet() != IndexSet(T_ax={0})
            True

        The practical implications of this imprecision are limited
        since we rarely need to :func:`gather` along a variable twice.

    :param value: The value to gather.
    :param IndexSet indexset: The :class:`IndexSet` of entries to select from ``value``.
    :param kwargs: Additional keyword arguments used by specific implementations.
    :return: A new value containing entries of ``value`` from ``indexset``.
    """
    raise NotImplementedError


@functools.singledispatch
def scatter(
    value, indexset: Optional[IndexSet] = None, *, result: Optional[T] = None, **kwargs
):
    """
    Assigns entries from an indexed value to entries in a larger indexed value.
    :func:`scatter` is primarily used internally in :class:`MultiWorldCounterfactual`
    for concisely and extensibly defining the semantics of counterfactuals.

    It also satisfies some equations with :func:`gather` and :func:`indices_of`
    that are useful for testing and debugging.

    Like :func:`torch.scatter` and assignment in term rewriting,
    :func:`scatter` is defined extensionally, meaning that values
    are treated as constant functions of variables not in their support.

    .. note::

        :func:`scatter` can be extended to new value types by registering
        an implementation for the type using :func:`functools.singledispatch` .

    :param value: The value to scatter.
    :param IndexSet indexset: The :class:`IndexSet` of entries of ``result`` to fill.
    :param Optional[T] result: The result to store the scattered value in.
    :param kwargs: Additional keyword arguments used by specific implementations.
    :return: The ``result``, with ``value`` scattered into the indices in ``indexset``.
    """
    raise NotImplementedError


@pyro.poutine.runtime.effectful(type="scatter_n")
def scatter_n(values: Dict[IndexSet, T], *, result: Optional[T] = None, **kwargs):
    """
    Scatters a dictionary of disjoint masked values into a single value
    using repeated calls to :func:``scatter``.

    :param partitioned_values: A dictionary mapping index sets to values.
    :return: A single value.
    """
    assert len(values) > 0
    assert all(isinstance(k, IndexSet) for k in values)
    for indices, value in values.items():
        result = scatter(value, indices, result=result, **kwargs)
    return result


@functools.singledispatch
def cond(fst, snd, case: Optional[T] = None, **kwargs):
    """
    Selection operation that is the sum-type analogue of :func:`scatter`
    in the sense that where :func:`scatter` propagates both of its arguments,
    :func:`cond` propagates only one, depending on the value of a boolean ``case`` .

    For a given ``fst`` , ``snd`` , and ``case`` , :func:`cond` returns
    ``snd`` if the ``case`` is true, and ``fst`` otherwise,
    analogous to a Python conditional expression ``snd if case else fst`` .
    Unlike a Python conditional expression, however, the case may be a tensor,
    and both branches are evaluated, as with :func:`torch.where` ::

        >>> fst, snd = torch.randn(2, 3), torch.randn(2, 3)
        >>> case = (fst < snd).all(-1)
        >>> x = cond(fst, snd, case, event_dim=1)
        >>> assert (x == torch.where(case[..., None], snd, fst)).all()

    .. note::

        :func:`cond` can be extended to new value types by registering
        an implementation for the type using :func:`functools.singledispatch` .

    :param fst: The value to return if ``case`` is ``False`` .
    :param snd: The value to return if ``case`` is ``True`` .
    :param case: A boolean value or tensor. If a tensor, should have event shape ``()`` .
    :param kwargs: Additional keyword arguments used by specific implementations.
    """
    raise NotImplementedError(f"cond not implemented for {type(fst)}")


@pyro.poutine.runtime.effectful(type="cond_n")
def cond_n(values: Dict[IndexSet, T], case: Union[bool, torch.Tensor], **kwargs):
    assert len(values) > 0
    assert all(isinstance(k, IndexSet) for k in values.keys())
    result: Optional[T] = None
    for indices, value in values.items():
        tst = torch.as_tensor(
            functools.reduce(
                operator.or_, [case == index for index in next(iter(indices.values()))]
            ),
            dtype=torch.bool,
        )
        result = cond(result if result is not None else value, value, tst, **kwargs)
    return result


@pyro.poutine.runtime.effectful(type="get_index_plates")
def get_index_plates() -> Dict[str, pyro.poutine.indep_messenger.CondIndepStackFrame]:
    return {}


def indexset_as_mask(
    indexset: IndexSet,
    *,
    event_dim: int = 0,
    name_to_dim_size: Optional[Dict[str, Tuple[int, int]]] = None,
    device: torch.device = torch.device("cpu"),
) -> torch.Tensor:
    """
    Get a dense mask tensor for indexing into a tensor from an indexset.
    """
    if name_to_dim_size is None:
        name_to_dim_size = {}
        for name, f in get_index_plates().items():
            assert f.dim is not None
            name_to_dim_size[name] = (f.dim, f.size)

    batch_shape = [1] * -min([dim for dim, _ in name_to_dim_size.values()], default=0)
    inds: List[Union[slice, torch.Tensor]] = [slice(None)] * len(batch_shape)
    for name, values in indexset.items():
        dim, size = name_to_dim_size[name]
        inds[dim] = torch.tensor(list(sorted(values)), dtype=torch.long)
        batch_shape[dim] = size
    mask = torch.zeros(tuple(batch_shape), dtype=torch.bool, device=device)
    mask[tuple(inds)] = True
    return mask[(...,) + (None,) * event_dim]



import functools
from typing import TypeVar

import pyro
import torch
from pyro.distributions.torch_distribution import TorchDistributionMixin

S = TypeVar("S")
T = TypeVar("T")


@functools.singledispatch
def uniform_proposal(
    support: pyro.distributions.constraints.Constraint, **kwargs
) -> TorchDistributionMixin:
    """
    This function heuristically constructs a probability distribution over a specified
    support. The choice of distribution depends on the type of support provided.

    - If the support is ``real``, it creates a wide Normal distribution
      and standard deviation, defaulting to ``(0,10)``.
    - If the support is ``boolean``, it creates a Bernoulli distribution with a fixed logit of ``0``,
      corresponding to success probability ``.5``.
    - If the support is an ``interval``, the transformed distribution is centered around the
      midpoint of the interval.

    :param support: The support used to create the probability distribution.
    :param kwargs: Additional keyword arguments.
    :return: A uniform probability distribution over the specified support.
    """
    if support is pyro.distributions.constraints.real:
        return pyro.distributions.Normal(0, 10).mask(False)
    elif support is pyro.distributions.constraints.boolean:
        return pyro.distributions.Bernoulli(logits=torch.zeros(()))
    else:
        tfm = pyro.distributions.transforms.biject_to(support)
        base = uniform_proposal(pyro.distributions.constraints.real, **kwargs)
        return pyro.distributions.TransformedDistribution(base, tfm)


@uniform_proposal.register
def _uniform_proposal_indep(
    support: pyro.distributions.constraints.independent,
    *,
    event_shape: torch.Size = torch.Size([]),
    **kwargs,
) -> TorchDistributionMixin:
    d = uniform_proposal(support.base_constraint, event_shape=event_shape, **kwargs)
    return d.expand(event_shape).to_event(support.reinterpreted_batch_ndims)


@uniform_proposal.register
def _uniform_proposal_integer(
    support: pyro.distributions.constraints.integer_interval,
    **kwargs,
) -> TorchDistributionMixin:
    if support.lower_bound != 0:
        raise NotImplementedError(
            "integer_interval with lower_bound > 0 not yet supported"
        )
    n = support.upper_bound - support.lower_bound + 1
    return pyro.distributions.Categorical(probs=torch.ones((n,)))



from typing import Any, Dict

import pyro

from chirho.indexed.ops import indices_of, union


class SpecifiedConditioningInferDict(pyro.poutine.runtime.InferDict):
    _specified_conditioning: bool


def site_is_ambiguous(msg: Dict[str, Any]) -> bool:
    """
    Helper function used with :func:`observe` to determine
    whether a site is observed or ambiguous.
    A sample site is ambiguous if it is marked observed, is downstream of an intervention,
    and the observed value's index variables are a strict subset of the distribution's
    indices and hence require clarification of which entries of the random variable
    are fixed/observed (as opposed to random/unobserved).
    """
    rv, obs = msg["args"][:2]
    value_indices = indices_of(obs, event_dim=len(rv.event_shape))
    dist_indices = indices_of(rv)
    return (
        bool(union(value_indices, dist_indices)) and value_indices != dist_indices
    ) or not msg["infer"].get("_specified_conditioning", True)


def no_ambiguity(msg: pyro.poutine.messenger.Message) -> SpecifiedConditioningInferDict:
    """
    Helper function used with :func:`pyro.poutine.infer_config` to inform
    :class:`FactualConditioningMessenger` that all ambiguity in the current
    context has been resolved.
    """
    return {"_specified_conditioning": True}



from typing import Any, Dict, Optional, Set, TypeVar

import pyro
import torch
from pyro.distributions.torch_distribution import TorchDistributionMixin

from chirho.indexed.handlers import DependentMaskMessenger, add_indices
from chirho.indexed.ops import IndexSet, gather, indexset_as_mask, scatter_n

T = TypeVar("T")


class CutModule(pyro.poutine.messenger.Messenger):
    """
    Converts a Pyro model into a module using the "cut" operation
    """

    vars: Set[str]

    def __init__(self, vars: Set[str]):
        self.vars = vars
        super().__init__()

    def _pyro_sample(self, msg: Dict[str, Any]) -> None:
        # There are 4 cases to consider for a sample site:
        # 1. The site appears in self.vars and is observed
        # 2. The site appears in self.vars and is not observed
        # 3. The site does not appear in self.vars and is observed
        # 4. The site does not appear in self.vars and is not observed
        if msg["name"] not in self.vars:
            if msg["is_observed"]:
                # use mask to remove the contribution of this observed site to the model log-joint
                msg["mask"] = (
                    msg["mask"] if msg["mask"] is not None else True
                ) & torch.tensor(False, dtype=torch.bool).expand(msg["fn"].batch_shape)
            else:
                pass

        # For sites that do not appear in module, rename them to avoid naming conflict
        if msg["name"] not in self.vars:
            msg["name"] = f"{msg['name']}_nuisance"


class CutComplementModule(pyro.poutine.messenger.Messenger):
    vars: Set[str]

    def __init__(self, vars: Set[str]):
        self.vars = vars
        super().__init__()

    def _pyro_sample(self, msg: Dict[str, Any]) -> None:
        # There are 4 cases to consider for a sample site:
        # 1. The site appears in self.vars and is observed
        # 2. The site appears in self.vars and is not observed
        # 3. The site does not appear in self.vars and is observed
        # 4. The site does not appear in self.vars and is not observed
        if msg["name"] in self.vars:
            # use mask to remove the contribution of this observed site to the model log-joint
            msg["mask"] = (
                msg["mask"] if msg["mask"] is not None else True
            ) & torch.tensor(False, dtype=torch.bool).expand(msg["fn"].batch_shape)


class SingleStageCut(DependentMaskMessenger):
    """
    Represent module and complement in a single Pyro model using plates
    """

    vars: Set[str]
    name: str

    def __init__(self, vars: Set[str], *, name: str = "__cut_plate"):
        self.vars = vars
        self.name = name
        super().__init__()

    def __enter__(self):
        add_indices(IndexSet(**{self.name: {0, 1}}))
        return super().__enter__()

    def get_mask(
        self,
        dist: pyro.distributions.Distribution,
        value: Optional[torch.Tensor],
        device: torch.device = torch.device("cpu"),
        name: Optional[str] = None,
    ) -> torch.Tensor:
        return indexset_as_mask(
            IndexSet(**{self.name: {0 if name in self.vars else 1}})
        )

    def _pyro_post_sample(self, msg: pyro.poutine.messenger.Message) -> None:
        if pyro.poutine.util.site_is_subsample(msg):
            return

        if (not msg["is_observed"]) and (msg["name"] in self.vars):
            assert isinstance(msg["fn"], TorchDistributionMixin)
            # discard the second value
            value_one = gather(
                msg["value"],
                IndexSet(**{self.name: {0}}),
                event_dim=msg["fn"].event_dim,
            )

            msg["value"] = scatter_n(
                {
                    IndexSet(**{self.name: {0}}): value_one,
                    IndexSet(**{self.name: {1}}): value_one.detach(),
                },
                event_dim=msg["fn"].event_dim,
            )



import numbers
from typing import Any, Dict, Optional, TypeVar, Union

import pyro
import pyro.infer.reparam
import torch
from pyro.distributions.torch_distribution import TorchDistributionMixin
from pyro.poutine.indep_messenger import CondIndepStackFrame, IndepMessenger

from chirho.indexed.ops import (
    IndexSet,
    cond,
    gather,
    get_index_plates,
    indices_of,
    scatter,
    union,
)

K = TypeVar("K")
T = TypeVar("T")


# Note that `gather` is defined using a `@functools.singledispatch` decorator,
# which in turn defines the `@gather.register` decorator used here
@gather.register
def _gather_number(
    value: numbers.Number,
    indexset: IndexSet,
    *,
    event_dim: Optional[int] = None,
    name_to_dim: Optional[Dict[str, int]] = None,
    **kwargs,
) -> Union[numbers.Number, torch.Tensor]:
    assert event_dim is None or event_dim == 0
    return gather(
        torch.as_tensor(value), indexset, event_dim=event_dim, name_to_dim=name_to_dim
    )


def _index_plate_dims() -> dict[str, int]:
    name_to_dim = {}
    for name, f in get_index_plates().items():
        assert f.dim is not None
        name_to_dim[name] = f.dim
    return name_to_dim


@gather.register
def _gather_tensor(
    value: torch.Tensor,
    indexset: IndexSet,
    *,
    event_dim: Optional[int] = None,
    name_to_dim: Optional[Dict[str, int]] = None,
    **kwargs,
) -> torch.Tensor:
    if event_dim is None:
        event_dim = 0

    if name_to_dim is None:
        name_to_dim = _index_plate_dims()

    result = value
    for name, indices in indexset.items():
        if name not in name_to_dim:
            continue
        dim = name_to_dim[name] - event_dim
        if len(result.shape) < -dim or result.shape[dim] == 1:
            continue
        result = result.index_select(
            name_to_dim[name] - event_dim,
            torch.tensor(list(sorted(indices)), device=value.device, dtype=torch.long),
        )
    return result


@gather.register(dict)
def _gather_dict(
    value: Dict[K, T], indices: IndexSet, *, event_dim: int = 0, **kwargs
) -> Dict[K, T]:
    return {
        k: gather(value[k], indices, event_dim=event_dim, **kwargs)
        for k in value.keys()
    }


@scatter.register
def _scatter_number(
    value: numbers.Number,
    indexset: IndexSet,
    *,
    result: Optional[torch.Tensor] = None,
    event_dim: Optional[int] = None,
    name_to_dim: Optional[Dict[str, int]] = None,
) -> Union[numbers.Number, torch.Tensor]:
    assert event_dim is None or event_dim == 0
    return scatter(
        torch.as_tensor(value),
        indexset,
        result=result,
        event_dim=event_dim,
        name_to_dim=name_to_dim,
    )


@scatter.register
def _scatter_tensor(
    value: torch.Tensor,
    indexset: IndexSet,
    *,
    result: Optional[torch.Tensor] = None,
    event_dim: Optional[int] = None,
    name_to_dim: Optional[Dict[str, int]] = None,
) -> torch.Tensor:
    if event_dim is None:
        event_dim = 0

    if name_to_dim is None:
        name_to_dim = _index_plate_dims()

    value = gather(value, indexset, event_dim=event_dim, name_to_dim=name_to_dim)
    indexset = union(
        indexset, indices_of(value, event_dim=event_dim, name_to_dim=name_to_dim)
    )

    if result is None:
        index_plates = get_index_plates()

        index_plate_dims = []
        for name, f in index_plates.items():
            assert f.dim is not None
            index_plate_dims.append(f.dim)

        result_shape = list(
            torch.broadcast_shapes(
                value.shape,
                (1,) * max([event_dim - dim for dim in index_plate_dims] + [0]),
            )
        )
        for name, indices in indexset.items():
            result_shape[name_to_dim[name] - event_dim] = index_plates[name].size
        result = value.new_zeros(result_shape)

    index = [
        torch.arange(0, result.shape[i], dtype=torch.long).reshape(
            (-1,) + (1,) * (len(result.shape) - 1 - i)
        )
        for i in range(len(result.shape))
    ]
    for name, indices in indexset.items():
        if result.shape[name_to_dim[name] - event_dim] > 1:
            index[name_to_dim[name] - event_dim] = torch.tensor(
                list(sorted(indices)), device=value.device, dtype=torch.long
            ).reshape((-1,) + (1,) * (event_dim - name_to_dim[name] - 1))

    result[tuple(index)] = value
    return result


@scatter.register(dict)
def _scatter_dict(
    value: Dict[K, T],
    indexset: IndexSet,
    *,
    result: Optional[Dict[K, Optional[T]]] = None,
    event_dim: Optional[int] = None,
    name_to_dim: Optional[Dict[str, int]] = None,
) -> Dict[K, Any]:

    if result is None:
        result = {k: None for k in value.keys()}

    for k in value.keys():
        result[k] = scatter(
            value[k],
            indexset,
            result=result[k],
            event_dim=event_dim,
            name_to_dim=name_to_dim,
        )

    return result


@indices_of.register
def _indices_of_number(value: numbers.Number, **kwargs) -> IndexSet:
    return IndexSet()


@indices_of.register
def _indices_of_bool(value: bool, **kwargs) -> IndexSet:
    return IndexSet()


@indices_of.register
def _indices_of_none(value: None, **kwargs) -> IndexSet:
    return IndexSet()


@indices_of.register
def _indices_of_tuple(value: tuple, **kwargs) -> IndexSet:
    if all(isinstance(v, int) for v in value):
        return indices_of(torch.Size(value), **kwargs)
    return union(*(indices_of(v, **kwargs) for v in value))


@indices_of.register
def _indices_of_shape(value: torch.Size, **kwargs) -> IndexSet:
    name_to_dim = (
        kwargs["name_to_dim"]
        if "name_to_dim" in kwargs
        else {name: f.dim for name, f in get_index_plates().items()}
    )
    value = value[: len(value) - kwargs.get("event_dim", 0)]
    return IndexSet(
        **{
            name: set(range(value[dim]))
            for name, dim in name_to_dim.items()
            if -dim <= len(value) and value[dim] > 1
        }
    )


@indices_of.register
def _indices_of_tensor(value: torch.Tensor, **kwargs) -> IndexSet:
    return indices_of(value.shape, **kwargs)


@indices_of.register
def _indices_of_distribution(value: TorchDistributionMixin, **kwargs) -> IndexSet:
    kwargs.pop("event_dim", None)
    return indices_of(value.batch_shape, event_dim=0, **kwargs)


@indices_of.register(dict)
def _indices_of_state(value: Dict[K, T], *, event_dim: int = 0, **kwargs) -> IndexSet:
    return union(
        *(indices_of(value[k], event_dim=event_dim, **kwargs) for k in value.keys())
    )


@cond.register(int)
@cond.register(float)
@cond.register(bool)
def _cond_number(
    fst: Union[bool, numbers.Number],
    snd: Union[bool, numbers.Number, torch.Tensor],
    case: Union[bool, torch.Tensor],
    **kwargs,
) -> torch.Tensor:
    return cond(
        torch.as_tensor(fst), torch.as_tensor(snd), torch.as_tensor(case), **kwargs
    )


@cond.register
def _cond_tensor(
    fst: torch.Tensor,
    snd: torch.Tensor,
    case: torch.Tensor,
    *,
    event_dim: int = 0,
    **kwargs,
) -> torch.Tensor:
    return torch.where(case[(...,) + (None,) * event_dim], snd, fst)


class _LazyPlateMessenger(IndepMessenger):
    prefix: str = "__index_plate__"

    def __init__(self, name: str, *args, **kwargs):
        self._orig_name: str = name
        super().__init__(f"{self.prefix}_{name}", *args, **kwargs)

    @property
    def frame(self) -> CondIndepStackFrame:
        return CondIndepStackFrame(
            name=self.name, dim=self.dim, size=self.size, counter=0
        )

    def _process_message(self, msg):
        if msg["type"] not in ("sample",) or pyro.poutine.util.site_is_subsample(msg):
            return
        if self._orig_name in union(
            indices_of(msg["value"], event_dim=msg["fn"].event_dim),
            indices_of(msg["fn"]),
        ):
            super()._process_message(msg)


def get_sample_msg_device(
    dist: TorchDistributionMixin,
    value: Optional[Union[torch.Tensor, float, int, bool]],
) -> torch.device:
    # some gross code to infer the device of the obs_mask tensor
    #   because distributions are hard to introspect
    if isinstance(value, torch.Tensor):
        return value.device
    else:
        dist_ = dist
        while hasattr(dist_, "base_dist"):
            dist_ = dist_.base_dist
        for param_name in dist_.arg_constraints.keys():
            p = getattr(dist_, param_name)
            if isinstance(p, torch.Tensor):
                return p.device
    raise ValueError(f"could not infer device for {dist} and {value}")


@pyro.poutine.runtime.effectful(type="add_indices")
def add_indices(indexset: IndexSet) -> IndexSet:
    return indexset
